#!/usr/bin/perl

# -----------------------------------------------------------------------------
# Name: OSFinger (osfinger.pl)
# Auth: Jason Bittel
# Date: 12/05/03
# Desc: A Perl implementation of TCP stack-based fingerprinting
#
# Implements TCP stack-based fingerprinting as described in:
# http://cert.uni-stuttgart.de/archive/bugtraq/2001/02/msg00195.html
# -----------------------------------------------------------------------------
use strict;
use Socket;
use Net::RawIP;
use Getopt::Std;
use vars qw($opt_b $opt_v $opt_h $opt_s $opt_d $opt_p $opt_o $opt_i);

# -----------------------------------------------------------------------------
# GLOBAL CONSTANTS
# -----------------------------------------------------------------------------
my $DST_PORT    = "80";   # Default destination port
my $INTERFACE   = "eth0"; # Default interface
my $NUM_PACKETS = 1;      # Number of packets to capture
my $PKT_SIZE    = 1500;   # Max packet size to capture
my $SRC_PORT    = int(rand 10000) + 1024;
                          # Random default source port (non-priviledged)
my $TIMEOUT     = 60;     # Packet capture timeout

# -----------------------------------------------------------------------------
# GLOBAL VARIABLES
# -----------------------------------------------------------------------------
my $batch;         # Batch mode flag
my $cap_desc;      # Packet capture descriptor
my $dst_address;   # Destination IP address
my $dst_port;      # Destination port
my $filter;        # Filter string for packet capture
my $interface;     # Interface name
my $offset;        # Packet offset
my $os_matchcount; # Number of OS's matched
my @os_matches;    # Names of matched OS's
my %packet;        # Returned data from remote host
my @parameters;    # Params to packet capture function
my $packet_pcap = new Net::RawIP;
                   # RawIP object
my $pid;           # Child PID
my $src_address;   # Source IP address
my $src_port;      # Source port
my $user;          # Current user id
my $verbose;       # Verbose flag

# -----------------------------------------------------------------------------
# Main Program
# -----------------------------------------------------------------------------

# Ensure user has proper permissions
$user = `id -u`;
chomp $user;
if ($user) # Not root (uid 0)
{
   print "\n You must have root access for this to work\n\n";
   exit;
}

# Process command line arguments
&get_arguments();

# Fork the program into a sending process and a listening/analysis process
$pid = fork();
if ($pid) # Parent process
{
   print "\n Initializing OSFinger...\n" unless $batch;
   &send_tcp_packet($src_address, $dst_address, $src_port, $dst_port);
   wait;
}
elsif($pid == 0) # Child process
{
   $filter   = "ip proto \\tcp and (dst host $src_address and dst port $src_port)";
   $cap_desc = $packet_pcap->pcapinit($interface,$filter,$PKT_SIZE,$TIMEOUT);
   $offset   = linkoffset($cap_desc);

   # Collect packet information
   loop($cap_desc,$NUM_PACKETS,\&get_tcp_packet,\@parameters);

   # Make sure port isn't closed
   if (&port_closed())
   {
      print "\n Port $dst_port is closed on $dst_address: try a different port\n\n" unless $batch;
      print "Host $dst_address: port $dst_port closed\n" if $batch;
      exit;
   }

   # Process packet information that was received
   &finger();
   &print_results();

   exit;
}

# -----------------------------------------------------------------------------
# Retrieve and process command line arguments
# -----------------------------------------------------------------------------
sub get_arguments
{
   getopts('bvhs:d:p:o:i:') or &print_usage();

   # Print help/usage information to the screen
   &print_usage() if ($opt_h);
   &print_usage() unless ($opt_d);

   # Copy command line arguments to internal variables
   $dst_address = $opt_d;
   $verbose     = 0 unless ($verbose = $opt_v);
   $batch       = 0 unless ($batch = $opt_b);
   $verbose     = 0 if ($batch); # Sanity check eliminating a batch/verbose combo
   $dst_port    = $DST_PORT unless ($dst_port = $opt_p);
   $src_port    = $SRC_PORT unless ($src_port = $opt_o);
   $interface   = $INTERFACE unless ($interface = $opt_i);
   $src_address = ${ifaddrlist()}{$interface} unless ($src_address = $opt_s);
}

# -----------------------------------------------------------------------------
# Craft a custom IP packet and send it
# -----------------------------------------------------------------------------
sub send_tcp_packet
{
   my $src_address = shift; # Source IP address
   my $dst_address = shift; # Destination IP address
   my $source      = shift; # Source port
   my $dest        = shift; # Destination port
   my $syn         = 1;     # SYN flag

   # Send a TCP SYN packet so that we can get a valid response
   my $packet = new Net::RawIP;
   $packet->set({ ip  => { saddr => $src_address, daddr => $dst_address  },
    	            tcp => { source => $source, dest => $dest, syn => $syn} });
   $packet->send;
   print "\n Sent > Source: $src_address:$source, Destination: $dst_address:$dest, SYN: $syn\n" if $verbose;
}

# -----------------------------------------------------------------------------
# Gather packets off the wire
# -----------------------------------------------------------------------------
sub get_tcp_packet
{
   my $curr_packet = $_[2]; # Current packet data (passed via @parameters)
   my $dnf;                 # Do not fragment bit
   my @pkt_data;            # Temporarily stores returned data

   # Ok, the theory here is to peel data out of the packet that will
   # allow us to match it to a known OS--additionally, we need the ACK,
   # RST, and SYN flags to know what kind of response was received
   $packet_pcap->bset($_[2], $offset);
   @pkt_data = $packet_pcap->get({ ip  => [qw(tos ttl saddr daddr)],
                                   tcp => [qw(source dest ack rst syn window)] });

   ############################################################################
   # Code for obtaining DNF bit obtained from passfingerprint.pl by Craig Smith
   $dnf = unpack("C", substr($curr_packet, $offset + 6, 1));
   $pkt_data[10] = (($dnf == 64) ? 1 : 0);
   ############################################################################

   # Map packet data into hash for legibility
   %packet = (
      "sourceip"   => inet_ntoa(pack("N",$pkt_data[2])), # Convert IP address to readable format
      "destip"     => inet_ntoa(pack("N",$pkt_data[3])), # Convert IP address to readable format
      "sourceport" => $pkt_data[4],
      "destport"   => $pkt_data[5],
      "tos"        => $pkt_data[0],
      "dnf"        => $pkt_data[10],
      "ttl"        => $pkt_data[1],
      "ack"        => $pkt_data[6],
      "rst"        => $pkt_data[7],
      "syn"        => $pkt_data[8],
      "window"     => $pkt_data[9],
   );

   print " Got  > TOS: $packet{'tos'}, DNF: $packet{'dnf'}, TTL: $packet{'ttl'}," if $verbose;
   print " ACK: $packet{'ack'}, RST: $packet{'rst'}, SYN: $packet{'syn'}," if $verbose;
   print " Win: $packet{'window'}\n" if $verbose;
}

# -----------------------------------------------------------------------------
# Analyze the received data to determine the OS--the core of the program
# -----------------------------------------------------------------------------
sub finger
{
   my @fingerprints; # Contents of fingerprint file
   my $fp;           # Current fingerprint
   my $packet_ttl;   # HPP for current packet
   my ($tos, $ttl, $win, $dnf, $os);
                     # Fields of fingerprint records

   # Read in fingerprint data from file for comparison
   open(FPFILE, "os_fingerprints") or die("\nError: Cannot open os_fingerprints - $!\n");
      @fingerprints = <FPFILE>;
   close(FPFILE);

   # Determine host path projection value
   $packet_ttl = &ttl_hpp($packet{"ttl"});

   print "\n Performing value matching for $packet{'window'} and $packet{'dnf'}\n" if $verbose;
   print " Mapped host path projection to $packet_ttl\n" if $verbose;
   foreach $fp (@fingerprints)
   {
      chomp $fp;
      next if $fp =~ /^\;/; # Skip comments
      next if $fp eq "";    # Skip blank lines

      ($tos, $ttl, $win, $dnf, $os) = split /,/,$fp;
      # First, we need to do a value match to see if it fits a known OS
      # We're not using TOS since some sources say it's app dependent
      if (($packet{"window"} == $win) && ($packet{"dnf"} == $dnf))
      {
         # Now we need to do the host path projection matching
         if ($packet_ttl == $ttl)
         {
            $os_matches[$os_matchcount] = $os;
            $os_matchcount++;
         }
      }
   }
}

# -----------------------------------------------------------------------------
# Check to see if we got a RST|ACK in response to determine port status
# -----------------------------------------------------------------------------
sub port_closed
{
   if (($packet{"rst"} == 1) && ($packet{"ack"} == 1)) # Port closed
   {
      return 1;
   }
   else # Port open
   {
      return 0;
   }
}

# -----------------------------------------------------------------------------
# Perform host path projection to find the initial TTL set by the host
# -----------------------------------------------------------------------------
sub ttl_hpp
{
   my $ttl = shift; # TTL of packet

   # NOTE: This HPP table in its text form was taken from
   # http://cert.uni-stuttgart.de/archive/bugtraq/2001/02/msg00195.html
   # +------------------------------------+
   # | TTL Value       |   TTL good match |
   # |------------------------------------|
   # |   32            |          0 -  32 |
   # |   64            |         33 -  64 |
   # |  128            |         65 - 128 |
   # |  255            |        129 - 255 |
   # +------------------------------------+

   # We're going to take the returned TTL value and map it into
   # its original TTL value according to the above table
   if (($ttl >= 0) && ($ttl <= 32))
   {
      return 32;
   }
   elsif (($ttl >= 33) && ($ttl <= 64))
   {
      return 64;
   }
   elsif (($ttl >= 65) && ($ttl <= 128))
   {
      return 128;
   }
   elsif (($ttl >= 129) && ($ttl <= 255))
   {
      return 255;
   }
}

# -----------------------------------------------------------------------------
# Print the results of the OS analysis
# -----------------------------------------------------------------------------
sub print_results
{
   if ($batch) # Do simple output for batch mode
   {
      if ($os_matchcount < 1) # No matches
      {
         print "Host $packet{'sourceip'}: no OS match found\n";
      }
      else
      {
         foreach (@os_matches) # At least one match found
         {
            print "Host $packet{'sourceip'}: $_\n";
         }
      }
   }
   else # Normal informational output
   {
      print "\n";
      if ($os_matchcount < 1) # No matches
      {
         print " Sorry, there were no OS matches for $packet{'sourceip'}\n\n";
         print " Here is the data that was returned:\n";
         print "        > TOS: $packet{'tos'}\n";
         print "        > TTL: $packet{'ttl'}\n";
         print "        > Win: $packet{'window'}\n";
         print "        > DNF: $packet{'dnf'}\n";
      }
      elsif ($os_matchcount == 1) # One match
      {
         print " There is $os_matchcount potential OS match for $packet{'sourceip'}:\n";
         print "        > $os_matches[0]\n";
      }
      else # More than one match
      {
         print " There are $os_matchcount potential OS matches for $packet{'sourceip'}:\n";
         foreach (@os_matches)
         {
            print "        > $_\n";
         }
      }
      print "\n";
   }
}

# -----------------------------------------------------------------------------
# Print usage/help information to the screen and exit
# -----------------------------------------------------------------------------
sub print_usage
{
   die <<USAGE;

 +--------------------------------------------------------------+
 |           OSFinger --- A Perl TSF OS Fingerprinter           |
 +--------------------------------------------------------------+

 Usage: $0 -d destination IP [-s source IP] [-p destination port]
    [-o source port ] [-i interface]

 By default, the following options are used if not specified: source IP is
 the current interface IP, destination port is $DST_PORT, source port is random
 (non-priviledged), and the ethernet interface is $INTERFACE. Obviously, these
 defaults can be overridden with the switches listed above.

 Additional Options:
    -b | Displays minimal output for running in batch mode
    -h | Prints this help/usage information to the screen
    -v | Displays verbose information during program operation

 Examples:
    $0 -d 192.168.0.5 -i eth1 -v
    $0 -d 216.239.41.99 -p 135 -o 6667

USAGE
}
